# -*- coding: utf-8 -*-
# Copyright 2023 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The Bgp_templates parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


UNIQUE_NEIB_ADD = "{{ nbr_address }}"


def _tmpl_maximum_prefix(config_data):
    conf = config_data.get("maximum_prefix", {})
    if conf:
        command = "maximum-prefix"
        if "max_limit" in conf:
            command += " " + str(conf["max_limit"])
        if "threshold_value" in conf:
            command += " " + str(conf["threshold_value"])
        if "restart" in conf:
            command += " restart " + str(conf["restart"])
        elif "warning_only" in conf:
            command += " warning-only"
        elif "discard_extra_paths" in conf:
            command += " discard-extra-paths"

    return command


def _tmpl_soft_reconfiguration(config_data):
    conf = config_data.get("soft_reconfiguration", {})
    if conf:
        command = "soft-reconfiguration "
        if "inbound" in conf:
            command += "inbound"
            if "set" in conf["inbound"]:
                pass
            elif "always" in conf["inbound"]:
                command += " always"
            if "inheritance_disable" in conf["inbound"]:
                command += " inheritance-disable"

    return command


def _tmpl_remove_private_AS(config_data):
    conf = config_data.get("remove_private_AS", {})
    if conf:
        command = "remove-private-AS"
        if "inbound" in conf:
            command += " inbound"
        if "entire_aspath" in conf:
            command += " entire-aspath"
        if "inheritance_disable" in conf:
            command += " inheritance-disable"
    return command


def _templ_local_as(config_data):
    conf = config_data.get("local_as", {})
    if conf.get("value"):
        command = "local-as " + str(conf.get("value", {}))
    if "no_prepend" in conf:
        command = "local-as"
        if "replace_as" in conf.get("no_prepend", {}):
            if "dual_as" in conf.get("no_prepend", {}).get("replace_as", {}):
                command += " no-prepend replace-as dual-as"
            elif "set" in conf.get("no_prepend", {}).get("replace_as", {}):
                command += " no-prepend replace-as"
        elif "set" in conf.get("no_prepend", {}):
            command += " no-prepend"
    return command


class Bgp_templatesTemplate(NetworkTemplate):
    def __init__(self, lines=None, module=None):
        super(Bgp_templatesTemplate, self).__init__(
            lines=lines,
            tmplt=self,
            module=module,
        )

    # fmt: off
    PARSERS = [
        {
            "name": "router",
            "getval": re.compile(
                r"""
                ^router\s
                bgp
                \s(?P<as_num>\S+)
                $""",
                re.VERBOSE,
            ),
            "setval": "router bgp {{ as_number }}",
            "compval": "as_number",
            "result": {"as_number": "{{ as_num }}"},
            "shared": True,
        },
        {
            "name": "address_family",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                (?P<address_family>\s+address-family\s(?P<afi>\S+)\s(?P<safi>\S+))
                $""", re.VERBOSE,
            ),
            "setval": "address-family {{ afi}} {{safi}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "afi": "{{ afi}}",
                                "safi": "{{safi}}",
                            },
                        },
                    },
                },
            },
            "shared": True,
        },
        {
            "name": "neighbor_group",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "neighbor-group {{ name}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "name": UNIQUE_NEIB_ADD,
                    },
                },
            },
            "shared": True,
        },
        {
            "name": "signalling",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sSignalling(?P<signalling>)
                (\sbgp\sdisable(?P<b_disable>))?
                (\sldp\sdisable(?P<l_disable>))?
                $""", re.VERBOSE,
            ),
            "setval": "{{ 'signalling bgp disable' if signalling.bgp_disable else 'signalling ldp disable' }} ",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "signalling": {
                                    "bgp_disable": "{{ True if b_disable is defined }}",
                                    "ldp_disable": "{{ True if l_disable is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "advertise.local_labeled_route.set",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                (?P<a_Set>\sadvertise\slocal-labeled-route(?P<a_set>))
                $""", re.VERBOSE,
            ),
            "setval": "advertise local-labeled-route",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "advertise": {
                                    "local_labeled_route": {
                                        "set": "{{ True if a_set is defined}}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "advertise.local_labeled_route.disable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sadvertise
                (\slocal-labeled-route\sdisable(?P<l_disable>))?
                $""", re.VERBOSE,
            ),
            "setval": "advertise local-labeled-route disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "advertise": {
                                    "local_labeled_route": {
                                        "disable": "{{ True if l_disable is defined}}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "advertise.permanent_network",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sadvertise
                (\spermanent-network(?P<set>))?
                $""", re.VERBOSE,
            ),
            "setval": "advertise permanent-network",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "advertise": {
                                    "permanent_network": "{{ True if set is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "aigp.set",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                (?P<aigp>\saigp)
                $""", re.VERBOSE,
            ),
            "setval": "aigp",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "aigp": {
                                    "set": "{{ True if aigp is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "aigp.disable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \saigp(?P<aigp>)
                (\sdisable(?P<disable>))
                $""", re.VERBOSE,
            ),
            "setval": "aigp disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "aigp": {
                                    "disable": "{{ True if disable is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "aigp.send_med",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \saigp(?P<aigp>)
                (\ssend\smed(?P<send_med>))?
                (\ssend\smed\sdisable(?P<send_disable>))?
                $""", re.VERBOSE,
            ),
            "setval": "'aigp send med disable' if {{aigp.send_med.disable}} is defined else 'aigp send med'",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "aigp": {
                                    "send_med": {
                                        "set": "{{ True if send_med is defined }}",
                                        "disable": "{{ True if send_disable is defined}}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "aigp.send_cost_community_disable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \saigp(?P<aigp>)
                (\ssend\scost-community\sdisable(?P<cc_disable>))?
                $""", re.VERBOSE,
            ),
            "setval": "aigp send cost-community disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "aigp": {
                                    "send_cost_community_disable": "{{True if cc_disable is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "allowas_in",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sallowas-in(?P<allowas_in>)(\s(?P<value>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": "allowas-in {{allowas_in.value if allowas_in.value is defined }}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "allowas_in": {
                                    "set": "{{True if allowas_in is defined and value is not defined}}",
                                    "value": "{{value }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "as_override",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sas-override(?P<as_override>)
                (\sinheritance-disable(?P<inheritance_disable>))?
                $""", re.VERBOSE,
            ),
            "setval": "as-override{{' inheritance-disable' if as_override.inheritance_disable is defined else ''}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "as_override": {
                                    "set": "{{True if as_override is defined "
                                           "and inheritance_disable is not defined}}",
                                    "inheritance_disable": "{{True if inheritance_disable is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "bestpath_origin_as_allow_invalid",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sbestpath\sorigin-as\sallow\sinvalid(?P<invalid>)
                $""", re.VERBOSE,
            ),
            "setval": "bestpath origin-as allow invalid",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "bestpath_origin_as_allow_invalid": "{{ True if invalid is defined}}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "capability_orf_prefix",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \scapability\sorf\sprefix\s(?P<capability_orf_prefix>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "capability orf prefix {{capability_orf_prefix }}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "capability_orf_prefix": "{{capability_orf_prefix}}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "default_originate.set",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s+default-originate(?P<default_originate>)
                $""", re.VERBOSE,
            ),
            "setval": "default-originate",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi}}': {
                                "default_originate": {
                                    "set": "{{True if default_originate is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "default_originate.route_policy",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s+default-originate(?P<default_originate>)
                (\sroute-policy\s(?P<route_policy>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": "default-originate route-policy {{default_originate.route_policy}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi}}': {
                                "default_originate": {
                                    "route_policy": "{{route_policy}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "default_originate.inheritance_disable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s+default-originate\s(?P<disable>inheritance-disable)
                $""", re.VERBOSE,
            ),
            "setval": "default-originate inheritance-disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi}}': {
                                "default_originate": {
                                    "inheritance_disable": "{{true if disable is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "encapsulation_type_srv6",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sencapsulation-type\ssrv6(?P<encapsulation_type_srv6>)
                $""", re.VERBOSE,
            ),
            "setval": "encapsulation-type srv6",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "encapsulation_type_srv6": "{{true if encapsulation_type_srv6 is defined}}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "long_lived_graceful_restart_capable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s+long-lived-graceful-restart
                \s(?P<capable>capable)
                $""", re.VERBOSE,
            ),
            "setval": "long-lived-graceful-restart capable",
            "compval": "long_lived_graceful_restart.capable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "long_lived_graceful_restart": {
                                    "capable": "{{True if capable is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "long_lived_graceful_restart_stale_time",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s+long-lived-graceful-restart
                \s+stale-time\ssend\s(?P<stale_time_send>\d+)\saccept\s(?P<accept>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "long-lived-graceful-restart stale-time send "
                      "{{stale_time.send}} accept {{stale_time.accept}}",
            "compval": "long_lived_graceful_restart.stale_time",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "long_lived_graceful_restart": {
                                    "stale_time": {
                                        "send": "{{stale_time_send}}",
                                        "accept": "{{accept}}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "maximum_prefix",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s+maximum-prefix
                (\s(?P<maximum_prefix>\d+))?
                (\s(?P<threshold_value>\d+))?
                (\srestart\s(?P<restart>\d+))?
                (\swarning-only\s(?P<warning_only>))?
                (\sdiscard-extra-paths\s(?P<discard_extra_paths>))?
                $""", re.VERBOSE,
            ),
            "setval": _tmpl_maximum_prefix,
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "maximum_prefix": {
                                    "max_limit": "{{maximum_prefix}}",
                                    "threshold_value": "{{threshold_value}}",
                                    "restart": "{{restart}}",
                                    "warning_only": "{{ True if warning_only is defined}}",
                                    "discard_extra_paths": "{{ True if discard_extra_paths is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },

        {
            "name": "multipath",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \smultipath(?P<multipath>)
                $""", re.VERBOSE,
            ),
            "setval": "multipath",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "multipath": "{{True if multipath is defined}}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "next_hop_self",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \snext-hop-self(?P<next_hop_self>)
                (\sinheritance-disable(?P<inheritance_disable>))?
                $""", re.VERBOSE,
            ),
            "setval": "next-hop-self{{' inheritance-disable' if next_hop_self.inheritance_disable is defined else ''}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "next_hop_self": {
                                    "set": "{{True if next_hop_self is defined and"
                                           " inheritance_disable is not defined}}",
                                    "inheritance_disable": "{{True if inheritance_disable is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "next_hop_unchanged.set",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \snext-hop-unchanged(?P<next_hop_unchanged>)
                $""", re.VERBOSE,
            ),
            "setval": "next-hop-unchanged",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi}}': {
                                "next_hop_unchanged": {
                                    "set": "{{True if next_hop_self is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "next_hop_unchanged.inheritance_disable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \snext-hop-unchanged(?P<next_hop_unchanged>)
                ((?P<inheritance_disable>)\sinheritance-disable)?
                $""", re.VERBOSE,
            ),
            "setval": "next-hop-unchanged inheritance-disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi}}': {
                                "next_hop_unchanged": {
                                    "inheritance_disable": "{{True if inheritance_disable is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "next_hop_unchanged.multipath",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \snext-hop-unchanged(?P<next_hop_unchanged>)
                (?P<multipath>\smultipath)?
                $""", re.VERBOSE,
            ),
            "setval": "next-hop-unchanged multipath",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi}}': {
                                "next_hop_unchanged": {
                                    "multipath": "{{True if multipath is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "optimal_route_reflection_group_name",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \soptimal-route-reflection\s(?P<group_name>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "optimal-route-reflection {{optimal_route_reflection_group_name}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "optimal_route_reflection_group_name": "{{ group_name}}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "orf_route_policy",
            "getval": re.compile(
                r"""
                \s+(?P<nbr_address>neighbor\s\S+)
                \sorf\sroute-policy\s(?P<orf_rr>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "orf route-policy {{orf_route_policy}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "orf_route_policy": "{{orf_rr}}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "origin_as",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sorigin-as\svalidation\sdisable(?P<origin_as>)
                $""", re.VERBOSE,
            ),
            "setval": "origin-as validation disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "origin_as": {
                                    "validation": {
                                        "disable": "{{True if origin_as is defined }}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "remove_private_AS.set",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sremove-private-AS(?P<remove_private_AS>)
                $""", re.VERBOSE,
            ),
            "setval": "remove-private-AS",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "remove_private_AS": {
                                    "set": "{{True if remove_private_AS is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "remove_private_AS",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sremove-private-AS(?P<remove_private_AS>)
                (\sinbound(?P<inbound>))?
                (\sentire-aspath(?P<entire_aspath>))?
                (\sinheritance-disable(?P<inheritance_disable>))?
                $""", re.VERBOSE,
            ),
            "setval": _tmpl_remove_private_AS,
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "remove_private_AS": {
                                    "inheritance_disable": "{{True if inheritance_disable is defined}}",
                                    "inbound": "{{True if inbound is defined}}",
                                    "entire_aspath": "{{True if entire_aspath is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "route_policy.inbound",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sroute-policy\s(?P<route_policy>\S+)
                \sin
                $""", re.VERBOSE,
            ),
            "setval": "route-policy {{route_policy.inbound}} in",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "route_policy": {
                                    "inbound": "{{route_policy}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "route_policy.outbound",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sroute-policy\s(?P<route_policy>\S+)
                \sout
                $""", re.VERBOSE,
            ),
            "setval": "route-policy {{route_policy.outbound}} out",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "route_policy": {
                                    "outbound": "{{route_policy}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "route_reflector_client",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sroute-reflector-client(?P<route_reflector_client>)
                (\sinheritance-disable(?P<inheritance_disable>))?
                $""", re.VERBOSE,
            ),
            "setval": "route-reflector-client{{' inheritance-disable' "
                      "if route_reflector_client.inheritance_disable is defined }}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "route_reflector_client": {
                                    "set": "{{True if route_reflector_client is defined and "
                                           "inheritance_disable is not defined }}",
                                    "inheritance_disable": "{{True if inheritance_disable is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "send_community_ebgp",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \ssend-community-ebgp(?P<send_community_ebgp>)
                (\sinheritance-disable(?P<inheritance_disable>))?
                $""", re.VERBOSE,
            ),
            "setval": "send-community-ebgp{{' inheritance-disable' "
                      "if send_community_ebgp.inheritance_disable is defined else ''}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "send_community_ebgp": {
                                    "set": "{{True if send_community_ebgp is defined and "
                                           "inheritance_disable is not defined}}",
                                    "inheritance_disable": "{{True if inheritance_disable is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "send_community_gshut_ebgp",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \ssend-community-gshut-ebgp(?P<send_community_gshut_ebg>)
                (\sinheritance-disable(?P<inheritance_disable>))?
                $""", re.VERBOSE,
            ),
            "setval": "send-community-gshut-ebgp{{' inheritance-disable' "
                      "if send_community_gshut_ebgp.inheritance_disable is defined else ''}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "send_community_gshut_ebgp": {
                                    "set": "{{True if send_community_gshut_ebg is defined and "
                                           "inheritance_disable is not defined}}",
                                    "inheritance_disable": "{{True if inheritance_disable is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "send_extended_community_ebgp",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \ssend-extended-community-ebgp(?P<send_extended_community_ebgp>)
                (\sinheritance-disable(?P<inheritance_disable>))?
                $""", re.VERBOSE,
            ),
            "setval": "send-extended-community-ebgp{{' inheritance-disable' "
                      "if send_extended_community_ebgp.inheritance_disable is defined else ''}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "send_extended_community_ebgp": {
                                    "set": "{{True if send_extended_community_ebgp is defined and "
                                           "inheritance_disable is not defined}}",
                                    "inheritance_disable": "{{True if inheritance_disable is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "send_multicast_attributes",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s+(?P<send_multicast_attributes>send-multicast-attributes)
                (\sdisable(?P<disable>))?
                $""", re.VERBOSE,
            ),
            "setval": "send-multicast-attributes{{' disable' "
                      "if send_multicast_attributes.disable is defined else ''}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "send_multicast_attributes": {
                                    "set": "{{True if send_multicast_attributes is "
                                           "defined and disable is not defined}}",
                                    "disable": "{{True if disable is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "soft_reconfiguration",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \ssoft-reconfiguration
                \sinbound(?P<inbound>)
                (\salways(?P<always>))?
                (\sinheritance-disable(?P<inheritance_disable>))?
                $""", re.VERBOSE,
            ),
            "setval": _tmpl_soft_reconfiguration,
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "soft_reconfiguration": {
                                    "inbound": {
                                        "set": "{{True if inbound is defined and "
                                               "inheritance_disable is not defined and "
                                               "always is not defined}}",
                                        "always": "{{True if always is defined }}",
                                        "inheritance_disable": "{{True if inheritance_disable is defined}}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "weight",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sweight\s(?P<weight>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "weight {{weight}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "weight": "{{weight}}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "use",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \suse\s(?P<af_use>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "weight {{weight}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "use": "{{af_use}}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "update.out_originator_loopcheck_set",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \supdate\sout\soriginator-loopcheck(?P<set>)
                $""", re.VERBOSE,
            ),
            "setval": "update out originator-loopcheck",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "update": {
                                    "out_originator_loopcheck_set": "{{True if set is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "update.out_originator_loopcheck_disable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \supdate\sout\soriginator-loopcheck(?P<set>)
                (\sdisable(?P<disable>))?
                $""", re.VERBOSE,
            ),
            "setval": "update out originator-loopcheck disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "address_family": {
                            '{{"address_family_" + afi + "_" + safi }}': {
                                "update": {
                                    "out_originator_loopcheck_disable": "{{True if disable is defined}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "advertisement_interval",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<advertise_in>advertisement-interval\s\d+)
                $""", re.VERBOSE,
            ),
            "setval": "advertisement-interval {{ advertisement_interval }}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "advertisement_interval": "{{ advertise_in.split(" ")[1] }}",
                    },
                },
            },
        },
        {
            "name": "bfd_fast_detect_disable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sbfd
                \sfast-detect
                \s(?P<disable>disable)
                $""", re.VERBOSE,
            ),
            "setval": "bfd fast-detect disable",
            "compval": "bfd.fast_detect.disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "bfd": {
                            "fast_detect": {"disable": "{{ True if disable is defined }}"},
                        },
                    },
                },
            },
        },
        {
            "name": "bfd_fast_detect_set",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sbfd
                \s(?P<fast_detect>fast-detect)
                $""", re.VERBOSE,
            ),
            "setval": "bfd fast-detect",
            "compval": "bfd.fast_detect.set",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "bfd": {
                            "fast_detect": {"set": "{{ True if fast_detect is defined }}"},
                        },
                    },
                },
            },
        },
        {
            "name": "bfd_fast_detect_strict_mode",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sbfd
                \sfast-detect
                \s(?P<strict_mode>strict-mode)
                $""", re.VERBOSE,
            ),
            "setval": "bfd fast-detect strict-mode",
            "compval": "bfd.fast_detect.strict_mode",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "bfd": {
                            "fast_detect": {"strict_mode": "{{ True if strict_mode is defined }}"},
                        },
                    },
                },
            },
        },
        {
            "name": "bfd_nbr_multiplier",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \sbfd
                \s(?P<multiplier>multiplier\s\S+)
                $""", re.VERBOSE,
            ),
            "setval": "bfd multiplier {{ bfd.multiplier}}",
            "compval": "bfd.multiplier",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD:
                    {
                        "bfd": {
                            "multiplier": "{{multiplier.split(" ")[1]}}",
                        },
                    },
                },
            },
        },
        {
            "name": "bfd_nbr_minimum_interval",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \sbfd
                \s(?P<min_interval>minimum-interval\s\S+)
                $""", re.VERBOSE,
            ),
            "setval": "bfd minimum-interval {{ bfd.minimum_interval}}",
            "compval": "bfd.minimum_interval",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "bfd": {
                            "minimum_interval": "{{min_interval.split(" ")[1]}}",
                        },
                    },
                },
            },
        },
        {
            "name": "bmp_activate",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sbmp-activate
                \s(?P<bmp_activate>server\s\d+)
                $""", re.VERBOSE,
            ),
            "setval": "bmp-activate server {{bmp_activate.server}}",
            "compval": "bmp_activate.serevr",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "bmp_activate": {"server": "{{ bmp_activate.split(" ")[1] }}"},
                    },
                },
            },
        },
        {
            "name": "neighbor_cluster_id",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<cluster_id>cluster-id\s\d+)
                $""", re.VERBOSE,
            ),
            "setval": "cluster-id {{ cluster_id }}",
            "compval": "cluster_id",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {"cluster_id": "{{ cluster_id.split(" ")[1] }}"},
                },
            },
        },
        {
            "name": "neighbor_description",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sdescription\s(?P<description>.+)
                $""", re.VERBOSE,
            ),
            "setval": "description {{ description }}",
            "compval": "description",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {"description": "{{ description }}"},
                },
            },
        },
        {
            "name": "dmz_link_bandwidth",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<dmz_link_bandwidth>dmz-link-bandwidth)
                $""", re.VERBOSE,
            ),
            "setval": "dmz-link-bandwidth",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "dmz_link_bandwidth": {
                            "set": "{{ True if dmz_link_bandwidth is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "dmz_link_bandwidth_inheritance_disable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sdmz-link-bandwidth
                \s(?P<dmz_link_bandwidth>inheritance_disable)
                $""", re.VERBOSE,
            ),
            "setval": "dmz-link-bandwidth inheritance-disable",
            "compval": "dmz_link_bandwidth.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "dmz_link_bandwidth": {
                            "inheritance_disable": "{{ True if dmz_link_bandwidth is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "dscp",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<dscp>dscp\s\S+)
                $""", re.VERBOSE,
            ),
            "setval": "dscp {{ dscp }}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "dscp": "{{ dscp.split(" ")[1] }}",
                    },
                },
            },
        },
        {
            "name": "ebgp_multihop_value",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<ebgp_multihop>ebgp-multihop\s\S+)
                $""", re.VERBOSE,
            ),
            "setval": "ebgp-multihop {{ ebgp_multihop.value}}",
            "compval": "ebgp_multihop.value",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "ebgp_multihop": {
                            "value": "{{ ebgp_multihop.split(" ")[1] }}",
                        },
                    },
                },
            },
        },
        {
            "name": "ebgp_multihop_mpls",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<ebgp_multihop>ebgp-multihop\s\S*\smpls)
                $""", re.VERBOSE,
            ),
            "setval": "ebgp-multihop mpls",
            "compval": "ebgp_multihop.mpls",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "ebgp_multihop": {"mpls": "{{ True if ebgp_multihop is defined }}"},
                    },
                },
            },
        },
        {
            "name": "ebgp_recv_extcommunity_dmz_set",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<ebgp_recv_extcommunity_dmz>ebgp-recv-extcommunity-dmz)
                $""", re.VERBOSE,
            ),
            "setval": "ebgp-recv-extcommunity-dmz inheritance-disable",
            "compval": "ebgp_recv_extcommunity_dm.set",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "ebgp_recv_extcommunity_dmz": {
                            "set": "{{ True if ebgp_recv_extcommunity_dmz is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "ebgp_recv_extcommunity_dmz",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<ebgp_recv_extcommunity_dmz>ebgp-recv-extcommunity-dmz\sinheritance-disable)
                $""", re.VERBOSE,
            ),
            "setval": "ebgp-recv-extcommunity-dmz inheritance-disable ",
            "compval": "ebgp_recv_extcommunity_dmz.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "ebgp_recv_extcommunity_dmz": {
                            "inheritance_disable": "{{ True if ebgp_recv_extcommunity_dmz is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "ebgp_send_extcommunity_dmz",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<ebgp_send_extcommunity_dmz>ebgp-send-extcommunity-dmz\sinheritance-disable)
                $""", re.VERBOSE,
            ),
            "setval": "ebgp-send-extcommunity-dmz inheritance-disable ",
            "compval": "ebgp_send_extcommunity_dmz.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "ebgp_send_extcommunity_dmz": {
                            "inheritance_disable": "{{ True if ebgp_send_extcommunity_dmz is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "ebgp_send_extcommunity_dmz_set",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<ebgp_send_extcommunity_dmz>ebgp-send-extcommunity-dmz)
                $""", re.VERBOSE,
            ),
            "setval": "ebgp-send-extcommunity-dmz",
            "compval": "ebgp_send_extcommunity_dmz.set",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "ebgp_send_extcommunity_dmz": {
                            "set": "{{ True if ebgp_send_extcommunity_dmz is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "ebgp_send_extcommunity_dmz_cumulatie",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<ebgp_send_extcommunity_dmz>ebgp-send-extcommunity-dmz\scumulatie)
                $""", re.VERBOSE,
            ),
            "setval": "ebgp-send-extcommunity-dmz cumulatie ",
            "compval": "ebgp_send_extcommunity_dmz.cumulatie",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "ebgp_send_extcommunity_dmz": {
                            "cumulatie": "{{ True if ebgp_send_extcommunity_dmz is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "egress_engineering",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<egress_engineering>egress-engineering\sinheritance-disable)
                $""", re.VERBOSE,
            ),
            "setval": "egress-engineering inheritance-disable ",
            "compval": "egress_engineering.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "egress_engineering": {
                            "inheritance_disable": "{{ True if egress_engineering is defined }}",
                        },
                    },
                },
            },

        },
        {
            "name": "egress_engineering_set",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<egress_engineering>egress-engineering)
                $""", re.VERBOSE,
            ),
            "setval": "egress-engineering",
            "compval": "egress_engineering.set",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "egress_engineering": {
                            "set": "{{ True if egress_engineering is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_enforce_first_as_disable",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<enforce_first_as_disable>enforce-first-as\sdisable)
                $""", re.VERBOSE,
            ),
            "setval": "enforce-first-as disable",
            "compval": "enforce_first_as.disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "enforce_first_as": {
                            "disable": "{{ True if enforce_first_as_disable is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_graceful_restart_restart_time",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<graceful_restart_restart_time>graceful-restart\srestart-time\s\d+)
                $""", re.VERBOSE,
            ),
            "setval": "graceful-restart restart-time {{ graceful_restart.restart_time}}",
            "compval": "graceful_restart.restart_time",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "graceful_restart": {
                            "restart_time": "{{ graceful_restart_restart_time.split(" ")[2] }}",
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_graceful_restart_stalepath_time",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<graceful_restart_stalepath_time>graceful-restart\sstalepath-time\s\d+)
                $""", re.VERBOSE,
            ),
            "setval": "graceful-restart stalepath-time {{ graceful_restart.stalepath_time}}",
            "compval": "graceful_restart.stalepath_time",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "graceful_restart": {
                            "stalepath_time": "{{ graceful_restart_stalepath_time.split(" ")[2] }}",
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_graceful_maintenance_set",
            "getval": re.compile(
                r"""
               \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<graceful_maintenance>graceful-maintenance)
                $""", re.VERBOSE,
            ),
            "setval": "graceful-maintenance",
            "compval": "graceful_maintenance.set",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "graceful_maintenance": {
                            "set": "{{ True if graceful_maintenance is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_graceful_maintenance_activate",
            "getval": re.compile(
                r"""
               \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<graceful_maintenance>graceful-maintenance\sactivate)
                $""", re.VERBOSE,
            ),
            "setval": "graceful-maintenance activate",
            "compval": "graceful_maintenance.activate.set",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "graceful_maintenance": {
                            "activate": {"set": "{{ True if graceful_maintenance is defined }}"},
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_graceful_maintenance_activate_inheritance_disable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<graceful_maintenance>activate\sinheritance-disable)
                $""", re.VERBOSE,
            ),
            "setval": "graceful-maintenance activate inheritance-disable",
            "compval": "graceful_maintenance.activate.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "graceful_maintenance": {
                            "activate": {
                                "inheritance_disable": "{{ True if graceful_maintenance is defined }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_graceful_maintenance_as_prepends",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<as_prepends>as-prepends\sinheritance-disable)
                $""", re.VERBOSE,
            ),
            "setval": "graceful-maintenance as-prepends inheritance-disable",
            "compval": "graceful_maintenance.as_prepends.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "graceful_maintenance": {
                            "as_prepends": {
                                "inheritance_disable": "{{ True if as_prepends is defined }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_graceful_maintenance_local_preference_disable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<local_preference>local-preference\sinheritance-disable)
                $""", re.VERBOSE,
            ),
            "setval": "graceful-maintenance local-preference inheritance-disable",
            "compval": "graceful_maintenance.local_preference.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "graceful_maintenance": {
                            "local_preference": {
                                "inheritance_disable": "{{ True if local_preference is defined }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_graceful_maintenance_local_preference",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<local_preference>local-preference\s\d+)
                $""", re.VERBOSE,
            ),
            "setval": "graceful-maintenance local-preference {{ graceful_maintenance.local_preference.value}}",
            "compval": "graceful_maintenance.local_preference.value",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "graceful_maintenance": {
                            "local_preference": {
                                "value": "{{ local_preference.split(" ")[1]}}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_graceful_maintenance_as_prepends_value",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<as_prepends>as-prepends\s\d+)
                $""", re.VERBOSE,
            ),
            "setval": "graceful-maintenance as-prepends {{ graceful_maintenance.as_prepends.value }}",
            "compval": "graceful_maintenance.as_prepends.value",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "graceful_maintenance": {
                            "as_prepends": {
                                "value": "{{ as_prepends.split(" ")[1]}}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ignore_connected_check_set",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<ignore_connected_check>ignore-connected-check)
                $""", re.VERBOSE,
            ),
            "setval": "ignore-connected-check",
            "compval": "ignore_connected_check.set",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "ignore_connected_check": {
                            "set": "{{ True if ignore_connected_check is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "ignore_connected_check",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<ignore_connected_check>ignore-connected-check\sinheritance-disable)
                $""", re.VERBOSE,
            ),
            "setval": "ignore-connected-check inheritance-disable ",
            "compval": "ignore_connected_check.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "ignore_connected_check": {
                            "inheritance_disable": "{{ True if ignore_connected_check is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "idle_watch_time",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \sidle-watch-time(?P<idle_watch_time>\s\S+)
                $""", re.VERBOSE,
            ),
            "setval": "idle-watch-time {{idle_watch_time}} ",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "idle_watch_time": "{{idle_watch_time}}",
                    },
                },
            },
        },
        {
            "name": "internal_vpn_client",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                (?P<internal_vpn_client>\sinternal-vpn-client)
                $""", re.VERBOSE,
            ),
            "setval": "internal-vpn-client",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "internal_vpn_client": "{{true if internal_vpn_client is defined}}",
                    },
                },
            },
        },
        {
            "name": "keychain",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<keychain>keychain\sinheritance-disable)
                $""", re.VERBOSE,
            ),
            "setval": "keychain inheritance-disable ",
            "compval": "keychain.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "keychain": {
                            "inheritance_disable": "{{ True if keychain is defined }}",
                        },
                    },
                },
            },

        },
        {
            "name": "keychain_name",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<keychain>keychain\s\S+)
                $""", re.VERBOSE,
            ),
            "setval": "keychain {{ keychain.name }}",
            "compval": "keychain.name",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "keychain": {
                            "name": "{{ keychain.split(" ")[1] }}",
                        },
                    },
                },
            },
        },
        {
            "name": "local_address",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \slocal
                \s(?P<local>address\sinheritance-disable)
                $""", re.VERBOSE,
            ),
            "setval": "local address inheritance-disable",
            "compval": "local.address.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "local": {
                            "address": {
                                "inheritance_disable": "{{ True if local is defined }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "local",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \slocal
                \s(?P<local>address\s\S+)
                $""", re.VERBOSE,
            ),
            "setval": "local address {{ local.address.ipv4_address }}",
            "compval": "local.address.ipv4_address",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "local": {
                            "address": {
                                "ipv4_address": "{{ local.split(" ")[1] }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "local_as_inheritance_disable",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<local_as>local-as\sinheritance-disable)
                $""", re.VERBOSE,
            ),
            "setval": "local-as inheritance-disable",
            "compval": "local_as.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "local_as": {
                            "inheritance_disable": "{{ True if local_as is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "local_as",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<local_as>local-as\s\S+)
                (\s(?P<no_prepend>no-prepend))?
                (\s(?P<replace_as>replace-as))?
                (\s(?P<dual_as>dual-as))?
                $""", re.VERBOSE,
            ),
            "setval": _templ_local_as,
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "local_as": {
                            "value": "{{ local_as.split(" ")[1] }}",
                            "no_prepend": {
                                "set": "{{ True if no_prepend is defined and replace_as is undefined and dual_as is undefined else None}}",
                                "replace_as": {
                                    "set": "{{ True if replace_as is defined and dual_as is undefined}}",
                                    "dual_as": "{{ not not dual_as}}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "local_address_subnet",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \slocal-address-subnet(?P<local>\s\S+)
                $""", re.VERBOSE,
            ),
            "setval": "local-address-subnet {{local_address_subnet}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "local_address_subnet": "{{local.split(" ")[1]}}",
                    },
                },
            },
        },
        {
            "name": "neighbor_log_message_in_value",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \slog
                \smessage
                \s(?P<value>in\s\d+)
                $""", re.VERBOSE,
            ),
            "setval": "log message in {{ log.message.in.value}}",
            "compval": "log.log_message.in.value",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "log": {
                            "log_message": {
                                "in": {
                                    "value": "{{ value.split(" ")[1] }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_log_message_in_disable",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \slog
                \smessage
                \s(?P<disable>in\sdisable)
                $""", re.VERBOSE,
            ),
            "setval": "log message in disable",
            "compval": "log.log_message.in.disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "log": {
                            "log_message": {
                                "in": {
                                    "disable": "{{ True if disable is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_log_message_in_inheritance_disable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \slog
                \smessage
                \s(?P<disable>in\sinheritance-diable)
                $""", re.VERBOSE,
            ),
            "setval": "log message in inheritance-diable",
            "compval": "log.log_message.in.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "log": {
                            "log_message": {
                                "in": {
                                    "inheritance_disable": "{{ True if disable is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_log_message_out_value",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \slog
                \smessage
                \s(?P<value>out\s\d+)
                $""", re.VERBOSE,
            ),
            "setval": "log message out {{ log.message.out.value}}",
            "compval": "log.log_message.out.value",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "log": {
                            "log_message": {
                                "out": {
                                    "value": "{{ value.split(" ")[1] }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_log_message_out_disable",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \slog
                \smessage
                \s(?P<disable>out\sdisable)
                $""", re.VERBOSE,
            ),
            "setval": "log message out disable",
            "compval": "log.log_message.out.disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "log": {
                            "log_message": {
                                "out": {
                                    "disable": "{{ True if disable is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_log_message_out_inheritance_disable",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \slog
                \smessage
                \s(?P<disable>out\sinheritance-diable)
                $""", re.VERBOSE,
            ),
            "setval": "log message out inheritance-diable",
            "compval": "log.log_message.out.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "log": {
                            "log_message": {
                                "out": {
                                    "inheritance_disable": "{{ True if disable is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "maximum_peers",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \smaximum-peers(?P<local>\s\S+)
                $""", re.VERBOSE,
            ),
            "setval": "maximum-peers {maximum_peers}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "maximum_peers": "{{local}}",
                    },
                },
            },

        },
        {
            "name": "password_inheritance_disable",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<password>password\sinheritance-disable)
                $""", re.VERBOSE,
            ),
            "setval": "password inheritance-disable",
            "compval": "password.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "password": {
                            "inheritance_disable": "{{ True if password is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "password_encrypted",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \spassword\sencrypted
                \s(?P<password>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "password encrypted {{password.encrypted}}",
            "compval": "password.encrypted",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "password": {
                            "encrypted": "{{ password }}",
                        },
                    },
                },
            },
        },
        {
            "name": "peer_set",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \speer-set(?P<local>\s\S+)
                $""", re.VERBOSE,
            ),
            "setval": "peer-set {peer_set}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "peer_set": "{{local}}",
                    },
                },
            },
        },
        {
            "name": "receive_buffer_size",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<receive_buffer_size>receive-buffer-size\s\d+)
                $""", re.VERBOSE,
            ),
            "setval": "receive-buffer-size {{ receive_buffer_size }}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "receive_buffer_size": "{{ receive_buffer_size.split(" ")[1] }}",
                    },
                },
            },
        },
        {
            "name": "send_buffer_size",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<send_buffer_size>send-buffer-size\s\d+)
                $""", re.VERBOSE,
            ),
            "setval": "send-buffer-size {{ send_buffer_size }}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "send_buffer_size": "{{ send_buffer_size.split(" ")[1] }}",
                    },
                },
            },
        },
        {
            "name": "precedence",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \sprecedence\s(?P<local>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "precedence {{precedence}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "precedence": "{{local}}",
                    },
                },
            },
        },
        {
            "name": "remote_as",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<remote_as>remote-as\s\S+)
                $""", re.VERBOSE,
            ),
            "setval": "remote-as {{ remote_as }}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "remote_as": "{{ remote_as.split(" ")[1] }}",
                    },
                },
            },
        },
        {
            "name": "remote_as_list",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<remote_as>remote-as-list\s\S+)
                $""", re.VERBOSE,
            ),
            "setval": "remote-as-list {{ remote_as_list }}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "remote_as_list": "{{ remote_as.split(" ")[1] }}",
                    },
                },
            },
        },
        {
            "name": "session_open_mode",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<session_open_mode>session-open-mode\s(active-only|both|passive-only))
                $""", re.VERBOSE,
            ),
            "setval": "session-open-mode {{ session_open_mode }}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "session_open_mode": "{{ session_open_mode.split(" ")[1] }}",
                    },
                },
            },
        },
        {
            "name": "neighbor_shutdown",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<shutdown>shutdown)
                $""", re.VERBOSE,
            ),
            "setval": "shutdown",
            "compval": "shutdown",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "shutdown": {
                            "set": "{{ True if shutdown is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_shutdown_inheritance_disable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<shutdown>shutdown\sinheritance_disable)
                $""", re.VERBOSE,
            ),
            "setval": "shutdown inheritance-disable",
            "compval": "shutdown.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "shutdown": {"inheritance_disable": "{{ True if shutdown is defined }}"},
                    },
                },
            },
        },
        {
            "name": "neighbor_tcp_mss_inheritance_disable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<tcp_mss_disable>tcp\smss\sinheritance-disable)
                $""", re.VERBOSE,
            ),
            "setval": "tcp mss inheritance-disable",
            "compval": "tcp.mss.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "tcp": {
                            "mss": {
                                "inheritance_disable": "{{ True if tcp_mss_disable is defined }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_tcp_mss",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<tcp_mss>tcp\smss\s\d+)
                $""", re.VERBOSE,
            ),
            "setval": "tcp mss {{ tcp.mss.value }}",
            "compval": "tcp.mss.value",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "tcp": {
                            "mss": {
                                "value": "{{ tcp_mss.split(" ")[2] }}",
                            },
                        },
                    },
                },
            },

        },
        {
            "name": "neighbor_timers_keepalive",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<timers_keepalive_time>timers\s\d+)
                \s(?P<timers_holdtime>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "timers {{ timers.keepalive_time}} {{ timers.holdtime }}",
            "compval": "timers",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "timers": {
                            "keepalive_time": "{{ timers_keepalive_time.split(" ")[1] }}",
                            "holdtime": "{{ timers_holdtime.split(" ")[0] }}",
                        },
                    },
                },
            },
        },
        {
            "name": "use.neighbor_group",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \suse\sneighbor-group\s(?P<neighbor_group>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "use neighbor-group {{ use.neighbor_group }}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "use": {
                            "neighbor_group": "{{ neighbor_group }}",
                        },
                    },
                },
            },

        },
        {
            "name": "use.session_group",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \suse\ssession-group\s(?P<session_group>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "use session-group {{ use.session_group }}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "use": {
                            "session_group": "{{ session_group }}",
                        },
                    },
                },
            },
        },
        {
            "name": "update_source",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \supdate-source
                \s(?P<update_source>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "update-source {{ update_source}}",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "update_source": "{{ update_source}}",
                    },
                },
            },
        },
        {
            "name": "neighbor_ttl_security_inheritance_disable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<ttl_security>ttl-security\sinheritance-disable)
                $""", re.VERBOSE,
            ),
            "setval": "ttl-security inheritance-disable",
            "compval": "ttl_security.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "ttl_security": {
                            "inheritance_disable": "{{ True if ttl_security is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_ttl_security",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<ttl_security>ttl-security)
                $""", re.VERBOSE,
            ),
            "setval": "ttl-security",
            "compval": "ttl_security.set",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "ttl_security": {
                            "set": "{{ True if ttl_security is defined }}",
                        },
                    },
                },
            },

        },
        {
            "name": "neighbor_update_in_filtering_attribute_filter_group",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<attribute_filter_group>attribute-filter\sgroup\s\S+)
                $""", re.VERBOSE,
            ),
            "setval": "update in filtering attribute-filter group {{ update.in.filtering.attribute_filter.group }}",
            "compval": "update.in.filtering.attribute_filter.group",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "update": {
                            "in": {
                                "filtering": {
                                    "attribute_filter": {
                                        "group": "{{ attribute_filter_group.split(" ")[2] }}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_update_in_filtering_logging_disable",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<logging_disable>logging\sdisable)
                $""", re.VERBOSE,
            ),
            "setval": "update in filtering logging disable",
            "compval": "update.in.filtering.logging.disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "update": {
                            "in": {
                                "filtering": {
                                    "logging": {
                                        "disable": "{{True if logging_disable is defined }}",
                                    },
                                },
                            },
                        },
                    },
                },
            },

        },
        {
            "name": "neighbor_update_in_filtering_message_buffers",
            "getval": re.compile(
                r"""
                 \s+neighbor-group\s(?P<nbr_address>\S+)
                \s(?P<message_buffers>message\sbuffers\s\d+)
                $""", re.VERBOSE,
            ),
            "setval": "update in filtering message buffers {{ update.in.filtering.message.buffers}}",
            "compval": "update.in.filtering.update_message.buffers",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "update": {
                            "in": {
                                "filtering": {
                                    "update_message": {
                                        "buffers": "{{ message_buffers.split(" ")[2] }}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_capability_additional_paths_send",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \scapability
                \sadditional-paths
                \s(?P<additional_paths_send>send)
                $""", re.VERBOSE,
            ),
            "setval": "capability additional-paths send",
            "compval": "capability.additional_paths.send.set",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "capability": {
                            "additional_paths": {
                                "send": {
                                    "set": "{{ True if additional_paths_send is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_capability_additional_paths_send_disable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \scapability
                \sadditional-paths
                \s(?P<additional_paths_send>send\sdisable)
                $""", re.VERBOSE,
            ),
            "setval": "capability additional-paths send disable",
            "compval": "capability.additional_paths.send.disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "capability": {
                            "additional_paths": {
                                "send": {
                                    "disable": "{{ True if additional_paths_send is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_capability_additional_paths_rcv",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \scapability
                \sadditional-paths
                \s(?P<additional_paths_receive>receive)
                $""", re.VERBOSE,
            ),
            "setval": "capability additional-paths receive",
            "compval": "capability.additional_paths.receive.set",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "capability": {
                            "additional_paths": {
                                "receive": {
                                    "set": "{{ True if additional_paths_receive is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_capability_additional_paths_rcv_disable",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \scapability
                \sadditional-paths
                \s(?P<additional_paths_receive_disable>receive\sdisable)
                $""", re.VERBOSE,
            ),
            "setval": "capability additional-paths receive disable",
            "compval": "capability.additional_paths.receive.disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "capability": {
                            "additional_paths": {
                                "receive": {
                                    "disable": "{{ True if additional_paths_receive_disable is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_capability_suppress_four_byte_AS",
            "getval": re.compile(
                r"""
               \s+neighbor-group\s(?P<nbr_address>\S+)
                \scapability
                \ssuppress
                \s(?P<suppress_4_byte_as>4-byte-as)
                $""", re.VERBOSE,
            ),
            "setval": "capability suppress 4-byte-as",
            "compval": "capability.suppress.four_byte_AS.set",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "capability": {
                            "suppress": {
                                "four_byte_AS": {
                                    "set": "{{ True if suppress_4_byte_as is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_capability_suppress_all",
            "getval": re.compile(
                r"""
                \s+neighbor-group\s(?P<nbr_address>\S+)
                \scapability
                \ssuppress
                \s(?P<all>all)
                $""", re.VERBOSE,
            ),
            "setval": "capability suppress all",
            "compval": "capability.suppress.all.set",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "capability": {
                            "suppress": {
                                "all": {
                                    "set": "{{ True if all is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor_capability_suppress_all_inheritance_disable",
            "getval": re.compile(
                r"""
               \s+neighbor-group\s(?P<nbr_address>\S+)
                \scapability
                \ssuppress
                \s(?P<all>all\sinheritance-disable)
                $""", re.VERBOSE,
            ),
            "setval": "capability suppress all inheritance-disable",
            "compval": "capability.suppress.all.inheritance_disable",
            "result": {
                "neighbor": {
                    UNIQUE_NEIB_ADD: {
                        "capability": {
                            "suppress": {
                                "all": {
                                    "inheritance_disable": "{{ True if all is defined }}",
                                },
                            },
                        },
                    },
                },
            },

        },

    ]
    # fmt: on



#
# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The nxos_bgp_neighbor_address_family config file.
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to its desired end-state is
created.
"""
from copy import deepcopy

from ansible.module_utils.six import iteritems
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.resource_module import (
    ResourceModule,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    dict_merge,
)

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.facts.facts import Facts
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.rm_templates.bgp_neighbor_address_family import (
    Bgp_neighbor_address_familyTemplate,
)


class Bgp_neighbor_address_family(ResourceModule):
    """
    The nxos_bgp_neighbor_address_family config class
    """

    def __init__(self, module):
        super(Bgp_neighbor_address_family, self).__init__(
            empty_fact_val={},
            facts_module=Facts(module),
            module=module,
            resource="bgp_neighbor_address_family",
            tmplt=Bgp_neighbor_address_familyTemplate(),
        )
        self.parsers = [
            "advertise_map.exist_map",
            "advertise_map.non_exist_map",
            "advertisement_interval",
            "allowas_in",
            "as_override",
            "capability.additional_paths.receive",
            "capability.additional_paths.send",
            "default_originate",
            "disable_peer_as_check",
            "filter_list.inbound",
            "filter_list.outbound",
            "inherit",
            "maximum_prefix",
            "next_hop_self",
            "next_hop_third_party",
            "prefix_list.inbound",
            "prefix_list.outbound",
            "rewrite_evpn_rt_asn",
            "rewrite_rt_asn",
            "route_map.inbound",
            "route_map.outbound",
            "route_reflector_client",
            "send_community.extended",
            "send_community.standard",
            "soft_reconfiguration_inbound",
            "soo",
            "suppress_inactive",
            "unsuppress_map",
            "weight",
        ]

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        if self.state not in ["parsed", "gathered"]:
            self.generate_commands()
            self.run_commands()
        return self.result

    def generate_commands(self):
        """Generate configuration commands to send based on
        want, have and desired state.
        """
        wantd = deepcopy(self.want)
        haved = deepcopy(self.have)

        for entry in wantd, haved:
            self._bgp_list_to_dict(entry)

        # if state is merged, merge want onto have and then compare
        if self.state == "merged":
            wantd = dict_merge(haved, wantd)

        # if state is deleted, empty out wantd and set haved to elements to delete
        if self.state == "deleted":
            if wantd:
                to_del = {
                    "neighbors": self._set_to_delete(haved, wantd),
                    "vrfs": {},
                }

                for k, hvrf in iteritems(haved.get("vrfs", {})):
                    wvrf = wantd.get("vrfs", {}).get(k, {})
                    to_del["vrfs"][k] = {
                        "neighbors": self._set_to_delete(hvrf, wvrf),
                        "vrf": k,
                    }
                haved.update(to_del)

            wantd = {}

        self._compare(want=wantd, have=haved)

        if self.commands:
            self.commands.insert(0, "router bgp {as_number}".format(**haved or wantd))

    def _compare(self, want, have, vrf=""):
        """Leverages the base class `compare()` method and
        populates the list of commands to be run by comparing
        the `want` and `have` data with the `parsers` defined
        for the Bgp_neighbor_address_family network resource.
        """
        w_nbrs = want.get("neighbors", {})
        h_nbrs = have.get("neighbors", {})

        if vrf:
            begin_vrf = len(self.commands)

        for k, w_nbr in iteritems(w_nbrs):
            begin = len(self.commands)
            h_nbr = h_nbrs.pop(k, {})
            want_afs = w_nbr.get("address_family", {})
            have_afs = h_nbr.get("address_family", {})

            for k, want_af in iteritems(want_afs):
                begin_af = len(self.commands)
                have_af = have_afs.pop(k, {})

                # swap `both` and `set` for idempotence
                if "send_community" in want_af:
                    if want_af["send_community"].get("both"):
                        want_af["send_community"] = {
                            "extended": True,
                            "standard": True,
                        }
                    elif want_af["send_community"].get("set"):
                        want_af["send_community"].update({"standard": True})

                self.compare(parsers=self.parsers, want=want_af, have=have_af)

                if len(self.commands) != begin_af or (not have_af and want_af):
                    self.commands.insert(
                        begin_af,
                        self._tmplt.render(want_af, "address_family", False),
                    )

            # remove remaining items in have for replaced
            for k, have_af in iteritems(have_afs):
                self.addcmd(have_af, "address_family", True)

            if len(self.commands) != begin:
                self.commands.insert(begin, "neighbor {0}".format(w_nbr["neighbor_address"]))

        if self.state in ["overridden", "deleted"]:
            for k, h_nbr in iteritems(h_nbrs):
                begin = len(self.commands)
                if not w_nbrs.pop(k, {}):
                    have_afs = h_nbr.get("address_family", {})
                    for k, have_af in iteritems(have_afs):
                        self.addcmd(have_af, "address_family", True)
                if len(self.commands) != begin:
                    self.commands.insert(begin, "neighbor {0}".format(h_nbr["neighbor_address"]))

        if vrf:
            if len(self.commands) != begin_vrf:
                self.commands.insert(begin_vrf, "vrf {0}".format(vrf))
        else:
            self._vrfs_compare(want, have)

    def _vrfs_compare(self, want, have):
        wvrfs = want.get("vrfs", {})
        hvrfs = have.get("vrfs", {})
        for k, wvrf in iteritems(wvrfs):
            h_vrf = hvrfs.pop(k, {})
            self._compare(want=wvrf, have=h_vrf, vrf=k)
        # remove remaining items in have
        for k, h_vrf in iteritems(hvrfs):
            self._compare(want={}, have=h_vrf, vrf=k)

    def _bgp_list_to_dict(self, data):
        if "neighbors" in data:
            for nbr in data["neighbors"]:
                if "address_family" in nbr:
                    nbr["address_family"] = {
                        (x["afi"], x.get("safi")): x for x in nbr["address_family"]
                    }
            data["neighbors"] = {x["neighbor_address"]: x for x in data["neighbors"]}

        if "vrfs" in data:
            for vrf in data["vrfs"]:
                self._bgp_list_to_dict(vrf)
            data["vrfs"] = {x["vrf"]: x for x in data["vrfs"]}

    def _set_to_delete(self, haved, wantd):
        neighbors = {}
        h_nbrs = haved.get("neighbors", {})
        w_nbrs = wantd.get("neighbors", {})

        for k, h_nbr in iteritems(h_nbrs):
            w_nbr = w_nbrs.pop(k, {})
            if w_nbr:
                neighbors[k] = h_nbr
                afs_to_del = {}
                h_addrs = h_nbr.get("address_family", {})
                w_addrs = w_nbr.get("address_family", {})
                for af, h_addr in iteritems(h_addrs):
                    if af in w_addrs:
                        afs_to_del[af] = h_addr
                neighbors[k]["address_family"] = afs_to_del

        return neighbors



# -*- coding: utf-8 -*-

# Copyright: (c) 2014, Will Thames <will@thames.id.au>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)


class ModuleDocFragment:
    # Common configuration for all AWS services
    # Note: If you're updating MODULES, PLUGINS probably needs updating too.

    # Formatted for Modules
    # - modules don't support 'env'
    MODULES = r"""
options:
  access_key:
    description:
      - AWS access key ID.
      - See the AWS documentation for more information about access tokens
        U(https://docs.aws.amazon.com/general/latest/gr/aws-sec-cred-types.html#access-keys-and-secret-access-keys).
      - The C(AWS_ACCESS_KEY_ID), C(AWS_ACCESS_KEY) or C(EC2_ACCESS_KEY)
        environment variables may also be used in decreasing order of
        preference.
      - The I(aws_access_key) and I(profile) options are mutually exclusive.
      - The I(aws_access_key_id) alias was added in release 5.1.0 for
        consistency with the AWS botocore SDK.
      - The I(ec2_access_key) alias has been deprecated and will be removed in a
        release after 2024-12-01.
      - Support for the C(EC2_ACCESS_KEY) environment variable has been
        deprecated and will be removed in a release after 2024-12-01.
    type: str
    aliases: ['aws_access_key_id', 'aws_access_key', 'ec2_access_key']
  secret_key:
    description:
      - AWS secret access key.
      - See the AWS documentation for more information about access tokens
        U(https://docs.aws.amazon.com/general/latest/gr/aws-sec-cred-types.html#access-keys-and-secret-access-keys).
      - The C(AWS_SECRET_ACCESS_KEY), C(AWS_SECRET_KEY), or C(EC2_SECRET_KEY)
        environment variables may also be used in decreasing order of
        preference.
      - The I(secret_key) and I(profile) options are mutually exclusive.
      - The I(aws_secret_access_key) alias was added in release 5.1.0 for
        consistency with the AWS botocore SDK.
      - The I(ec2_secret_key) alias has been deprecated and will be removed in a
        release after 2024-12-01.
      - Support for the C(EC2_SECRET_KEY) environment variable has been
        deprecated and will be removed in a release after 2024-12-01.
    type: str
    aliases: ['aws_secret_access_key', 'aws_secret_key', 'ec2_secret_key']
  session_token:
    description:
      - AWS STS session token for use with temporary credentials.
      - See the AWS documentation for more information about access tokens
        U(https://docs.aws.amazon.com/general/latest/gr/aws-sec-cred-types.html#access-keys-and-secret-access-keys).
      - The C(AWS_SESSION_TOKEN), C(AWS_SECURITY_TOKEN) or C(EC2_SECURITY_TOKEN)
        environment variables may also be used in decreasing order of preference.
      - The I(security_token) and I(profile) options are mutually exclusive.
      - Aliases I(aws_session_token) and I(session_token) were added in release
        3.2.0, with the parameter being renamed from I(security_token) to
        I(session_token) in release 6.0.0.
      - The I(security_token), I(aws_security_token), and I(access_token)
        aliases have been deprecated and will be removed in a release after
        2024-12-01.
      - Support for the C(EC2_SECRET_KEY) and C(AWS_SECURITY_TOKEN) environment
        variables has been deprecated and will be removed in a release after
        2024-12-01.
    type: str
    aliases: ['aws_session_token', 'security_token', 'aws_security_token', 'access_token']
  profile:
    description:
      - A named AWS profile to use for authentication.
      - See the AWS documentation for more information about named profiles
        U(https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-profiles.html).
      - The C(AWS_PROFILE) environment variable may also be used.
      - The I(profile) option is mutually exclusive with the I(aws_access_key),
        I(aws_secret_key) and I(security_token) options.
    type: str
    aliases: ['aws_profile']

  endpoint_url:
    description:
      - URL to connect to instead of the default AWS endpoints.  While this
        can be used to connection to other AWS-compatible services the
        amazon.aws and community.aws collections are only tested against
        AWS.
      - The  C(AWS_URL) or C(EC2_URL) environment variables may also be used,
        in decreasing order of preference.
      - The I(ec2_url) and I(s3_url) aliases have been deprecated and will be
        removed in a release after 2024-12-01.
      - Support for the C(EC2_URL) environment variable has been deprecated and
        will be removed in a release after 2024-12-01.
    type: str
    aliases: ['ec2_url', 'aws_endpoint_url', 's3_url' ]

  aws_ca_bundle:
    description:
      - The location of a CA Bundle to use when validating SSL certificates.
      - The C(AWS_CA_BUNDLE) environment variable may also be used.
    type: path
  validate_certs:
    description:
      - When set to C(false), SSL certificates will not be validated for
        communication with the AWS APIs.
      - Setting I(validate_certs=false) is strongly discouraged, as an
        alternative, consider setting I(aws_ca_bundle) instead.
    type: bool
    default: true
  aws_config:
    description:
      - A dictionary to modify the botocore configuration.
      - Parameters can be found in the AWS documentation
        U(https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html#botocore.config.Config).
    type: dict
  debug_botocore_endpoint_logs:
    description:
      - Use a C(botocore.endpoint) logger to parse the unique (rather than total)
        C("resource:action") API calls made during a task, outputing the set to
        the resource_actions key in the task results. Use the
        C(aws_resource_action) callback to output to total list made during
        a playbook.
      - The C(ANSIBLE_DEBUG_BOTOCORE_LOGS) environment variable may also be used.
    type: bool
    default: false
notes:
  - B(Caution:) For modules, environment variables and configuration files are
    read from the Ansible 'host' context and not the 'controller' context.
    As such, files may need to be explicitly copied to the 'host'.  For lookup
    and connection plugins, environment variables and configuration files are
    read from the Ansible 'controller' context and not the 'host' context.
  - The AWS SDK (boto3) that Ansible uses may also read defaults for credentials
    and other settings, such as the region, from its configuration files in the
    Ansible 'host' context (typically C(~/.aws/credentials)).
    See U(https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html)
    for more information.
"""

    # Formatted for non-module plugins
    # - modules don't support 'env'
    PLUGINS = r"""
options:
  access_key:
    description:
      - AWS access key ID.
      - See the AWS documentation for more information about access tokens
        U(https://docs.aws.amazon.com/general/latest/gr/aws-sec-cred-types.html#access-keys-and-secret-access-keys).
      - The I(aws_access_key) and I(profile) options are mutually exclusive.
      - The I(aws_access_key_id) alias was added in release 5.1.0 for
        consistency with the AWS botocore SDK.
      - The I(ec2_access_key) alias has been deprecated and will be removed in a
        release after 2024-12-01.
    type: str
    aliases: ['aws_access_key_id', 'aws_access_key', 'ec2_access_key']
    env:
      - name: AWS_ACCESS_KEY_ID
      - name: AWS_ACCESS_KEY
      - name: EC2_ACCESS_KEY
        deprecated:
          removed_at_date: '2024-12-01'
          collection_name: amazon.aws
          why: 'EC2 in the name implied it was limited to EC2 resources.  However, it is used for all connections.'
          alternatives: AWS_ACCESS_KEY_ID
  secret_key:
    description:
      - AWS secret access key.
      - See the AWS documentation for more information about access tokens
        U(https://docs.aws.amazon.com/general/latest/gr/aws-sec-cred-types.html#access-keys-and-secret-access-keys).
      - The I(secret_key) and I(profile) options are mutually exclusive.
      - The I(aws_secret_access_key) alias was added in release 5.1.0 for
        consistency with the AWS botocore SDK.
      - The I(ec2_secret_key) alias has been deprecated and will be removed in a
        release after 2024-12-01.
    type: str
    aliases: ['aws_secret_access_key', 'aws_secret_key', 'ec2_secret_key']
    env:
      - name: AWS_SECRET_ACCESS_KEY
      - name: AWS_SECRET_KEY
      - name: EC2_SECRET_KEY
        deprecated:
          removed_at_date: '2024-12-01'
          collection_name: amazon.aws
          why: 'EC2 in the name implied it was limited to EC2 resources.  However, it is used for all connections.'
          alternatives: AWS_SECRET_ACCESS_KEY
  session_token:
    description:
      - AWS STS session token for use with temporary credentials.
      - See the AWS documentation for more information about access tokens
        U(https://docs.aws.amazon.com/general/latest/gr/aws-sec-cred-types.html#access-keys-and-secret-access-keys).
      - The I(security_token) and I(profile) options are mutually exclusive.
      - Aliases I(aws_session_token) and I(session_token) were added in release
        3.2.0, with the parameter being renamed from I(security_token) to
        I(session_token) in release 6.0.0.
      - The I(security_token), I(aws_security_token), and I(access_token)
        aliases have been deprecated and will be removed in a release after
        2024-12-01.
    type: str
    aliases: ['aws_session_token', 'security_token', 'aws_security_token', 'access_token']
    env:
      - name: AWS_SESSION_TOKEN
      - name: AWS_SECURITY_TOKEN
        deprecated:
          removed_at_date: '2024-12-01'
          collection_name: amazon.aws
          why: 'AWS_SECURITY_TOKEN was used for compatibility with the original boto SDK, support for which has been dropped'
          alternatives: AWS_SESSION_TOKEN
      - name: EC2_SECURITY_TOKEN
        deprecated:
          removed_at_date: '2024-12-01'
          collection_name: amazon.aws
          why: 'EC2 in the name implied it was limited to EC2 resources.  However, it is used for all connections.'
          alternatives: AWS_SESSION_TOKEN

  profile:
    description:
      - A named AWS profile to use for authentication.
      - See the AWS documentation for more information about named profiles
        U(https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-profiles.html).
      - The I(profile) option is mutually exclusive with the I(aws_access_key),
        I(aws_secret_key) and I(security_token) options.
      - The I(boto_profile) alias has been deprecated and will be removed in a
        release after 2024-12-01.
    type: str
    aliases: ['aws_profile', 'boto_profile']
    env:
      - name: AWS_PROFILE
      - name: AWS_DEFAULT_PROFILE
  endpoint_url:
    description:
      - URL to connect to instead of the default AWS endpoints.  While this
        can be used to connection to other AWS-compatible services the
        amazon.aws and community.aws collections are only tested against
        AWS.
      - The I(endpoint) alias has been deprecated and will be
        removed in a release after 2024-12-01.
    type: str
    aliases: ['aws_endpoint_url', 'endpoint' ]
    env:
      - name: AWS_URL
      - name: EC2_URL
        deprecated:
          removed_at_date: '2024-12-01'
          collection_name: amazon.aws
          why: 'EC2 in the name implied it was limited to EC2 resources.  However, it is used for all connections.'
          alternatives: AWS_URL

notes:
  - B(Caution:) For modules, environment variables and configuration files are
    read from the Ansible 'host' context and not the 'controller' context.
    As such, files may need to be explicitly copied to the 'host'.  For lookup
    and connection plugins, environment variables and configuration files are
    read from the Ansible 'controller' context and not the 'host' context.
  - The AWS SDK (boto3) that Ansible uses may also read defaults for credentials
    and other settings, such as the region, from its configuration files in the
    Ansible 'host' context (typically C(~/.aws/credentials)).
    See U(https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html)
    for more information.
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2020, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = r'''
---
module: x509_crl_info
version_added: '1.0.0'
short_description: Retrieve information on Certificate Revocation Lists (CRLs)
description:
    - This module allows one to retrieve information on Certificate Revocation Lists (CRLs).
requirements:
    - cryptography >= 1.2
author:
    - Felix Fontein (@felixfontein)
extends_documentation_fragment:
    - community.crypto.attributes
    - community.crypto.attributes.info_module
    - community.crypto.name_encoding
options:
    path:
        description:
            - Remote absolute path where the generated CRL file should be created or is already located.
            - Either O(path) or O(content) must be specified, but not both.
        type: path
    content:
        description:
            - Content of the X.509 CRL in PEM format, or Base64-encoded X.509 CRL.
            - Either O(path) or O(content) must be specified, but not both.
        type: str
    list_revoked_certificates:
        description:
            - If set to V(false), the list of revoked certificates is not included in the result.
            - This is useful when retrieving information on large CRL files. Enumerating all revoked
              certificates can take some time, including serializing the result as JSON, sending it to
              the Ansible controller, and decoding it again.
        type: bool
        default: true
        version_added: 1.7.0

notes:
    - All timestamp values are provided in ASN.1 TIME format, in other words, following the C(YYYYMMDDHHMMSSZ) pattern.
      They are all in UTC.
seealso:
    - module: community.crypto.x509_crl
    - plugin: community.crypto.x509_crl_info
      plugin_type: filter
      description: A filter variant of this module.
    - plugin: community.crypto.to_serial
      plugin_type: filter
'''

EXAMPLES = r'''
- name: Get information on CRL
  community.crypto.x509_crl_info:
    path: /etc/ssl/my-ca.crl
  register: result

- name: Print the information
  ansible.builtin.debug:
    msg: "{{ result }}"

- name: Get information on CRL without list of revoked certificates
  community.crypto.x509_crl_info:
    path: /etc/ssl/very-large.crl
    list_revoked_certificates: false
  register: result
'''

RETURN = r'''
format:
    description:
        - Whether the CRL is in PEM format (V(pem)) or in DER format (V(der)).
    returned: success
    type: str
    sample: pem
    choices:
        - pem
        - der
issuer:
    description:
        - The CRL's issuer.
        - Note that for repeated values, only the last one will be returned.
        - See O(name_encoding) for how IDNs are handled.
    returned: success
    type: dict
    sample: {"organizationName": "Ansible", "commonName": "ca.example.com"}
issuer_ordered:
    description: The CRL's issuer as an ordered list of tuples.
    returned: success
    type: list
    elements: list
    sample: [["organizationName", "Ansible"], ["commonName": "ca.example.com"]]
last_update:
    description: The point in time from which this CRL can be trusted as ASN.1 TIME.
    returned: success
    type: str
    sample: '20190413202428Z'
next_update:
    description: The point in time from which a new CRL will be issued and the client has to check for it as ASN.1 TIME.
    returned: success
    type: str
    sample: '20190413202428Z'
digest:
    description: The signature algorithm used to sign the CRL.
    returned: success
    type: str
    sample: sha256WithRSAEncryption
revoked_certificates:
    description: List of certificates to be revoked.
    returned: success if O(list_revoked_certificates=true)
    type: list
    elements: dict
    contains:
        serial_number:
            description:
                - Serial number of the certificate.
                - This return value is an B(integer). If you need the serial numbers as a colon-separated hex string,
                  such as C(11:22:33), you need to convert it to that form with P(community.crypto.to_serial#filter).
            type: int
            sample: 1234
        revocation_date:
            description: The point in time the certificate was revoked as ASN.1 TIME.
            type: str
            sample: '20190413202428Z'
        issuer:
            description:
                - The certificate's issuer.
                - See O(name_encoding) for how IDNs are handled.
            type: list
            elements: str
            sample: ["DNS:ca.example.org"]
        issuer_critical:
            description: Whether the certificate issuer extension is critical.
            type: bool
            sample: false
        reason:
            description:
                - The value for the revocation reason extension.
            type: str
            sample: key_compromise
            choices:
                - unspecified
                - key_compromise
                - ca_compromise
                - affiliation_changed
                - superseded
                - cessation_of_operation
                - certificate_hold
                - privilege_withdrawn
                - aa_compromise
                - remove_from_crl
        reason_critical:
            description: Whether the revocation reason extension is critical.
            type: bool
            sample: false
        invalidity_date:
            description: |
                The point in time it was known/suspected that the private key was compromised
                or that the certificate otherwise became invalid as ASN.1 TIME.
            type: str
            sample: '20190413202428Z'
        invalidity_date_critical:
            description: Whether the invalidity date extension is critical.
            type: bool
            sample: false
'''


import base64
import binascii

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_native

from ansible_collections.community.crypto.plugins.module_utils.crypto.basic import (
    OpenSSLObjectError,
)

from ansible_collections.community.crypto.plugins.module_utils.crypto.pem import (
    identify_pem_format,
)

from ansible_collections.community.crypto.plugins.module_utils.crypto.module_backends.crl_info import (
    get_crl_info,
)


def main():
    module = AnsibleModule(
        argument_spec=dict(
            path=dict(type='path'),
            content=dict(type='str'),
            list_revoked_certificates=dict(type='bool', default=True),
            name_encoding=dict(type='str', default='ignore', choices=['ignore', 'idna', 'unicode']),
        ),
        required_one_of=(
            ['path', 'content'],
        ),
        mutually_exclusive=(
            ['path', 'content'],
        ),
        supports_check_mode=True,
    )

    if module.params['content'] is None:
        try:
            with open(module.params['path'], 'rb') as f:
                data = f.read()
        except (IOError, OSError) as e:
            module.fail_json(msg='Error while reading CRL file from disk: {0}'.format(e))
    else:
        data = module.params['content'].encode('utf-8')
        if not identify_pem_format(data):
            try:
                data = base64.b64decode(module.params['content'])
            except (binascii.Error, TypeError) as e:
                module.fail_json(msg='Error while Base64 decoding content: {0}'.format(e))

    try:
        result = get_crl_info(module, data, list_revoked_certificates=module.params['list_revoked_certificates'])
        module.exit_json(**result)
    except OpenSSLObjectError as e:
        module.fail_json(msg=to_native(e))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type
try:
    from dnacentersdk import api, exceptions
except ImportError:
    DNAC_SDK_IS_INSTALLED = False
else:
    DNAC_SDK_IS_INSTALLED = True
from ansible.module_utils._text import to_native
from ansible.module_utils.common import validation
from abc import ABCMeta, abstractmethod
try:
    import logging
except ImportError:
    LOGGING_IN_STANDARD = False
else:
    LOGGING_IN_STANDARD = True
import os.path
import copy
import json
# import datetime
import inspect
import re


class DnacBase():

    """Class contains members which can be reused for all intent modules"""

    __metaclass__ = ABCMeta
    __is_log_init = False

    def __init__(self, module):
        self.module = module
        self.params = module.params
        self.config = copy.deepcopy(module.params.get("config"))
        self.have = {}
        self.want = {}
        self.validated_config = []
        self.msg = ""
        self.status = "success"
        dnac_params = self.get_dnac_params(self.params)
        self.dnac = DNACSDK(params=dnac_params)
        self.dnac_apply = {'exec': self.dnac._exec}
        self.get_diff_state_apply = {'merged': self.get_diff_merged,
                                     'deleted': self.get_diff_deleted,
                                     'replaced': self.get_diff_replaced,
                                     'overridden': self.get_diff_overridden,
                                     'gathered': self.get_diff_gathered,
                                     'rendered': self.get_diff_rendered,
                                     'parsed': self.get_diff_parsed
                                     }
        self.verify_diff_state_apply = {'merged': self.verify_diff_merged,
                                        'deleted': self.verify_diff_deleted,
                                        'replaced': self.verify_diff_replaced,
                                        'overridden': self.verify_diff_overridden,
                                        'gathered': self.verify_diff_gathered,
                                        'rendered': self.verify_diff_rendered,
                                        'parsed': self.verify_diff_parsed
                                        }
        self.dnac_log = dnac_params.get("dnac_log")

        if self.dnac_log and not DnacBase.__is_log_init:
            self.dnac_log_level = dnac_params.get("dnac_log_level") or 'WARNING'
            self.dnac_log_level = self.dnac_log_level.upper()
            self.validate_dnac_log_level()
            self.dnac_log_file_path = dnac_params.get("dnac_log_file_path") or 'dnac.log'
            self.validate_dnac_log_file_path()
            self.dnac_log_mode = 'w' if not dnac_params.get("dnac_log_append") else 'a'
            self.setup_logger('logger')
            self.logger = logging.getLogger('logger')
            DnacBase.__is_log_init = True
            self.log('Logging configured and initiated', "DEBUG")
        elif not self.dnac_log:
            # If dnac_log is False, return an empty logger
            self.logger = logging.getLogger('empty_logger')

        self.log('Cisco Catalyst Center parameters: {0}'.format(dnac_params), "DEBUG")
        self.supported_states = ["merged", "deleted", "replaced", "overridden", "gathered", "rendered", "parsed"]
        self.result = {"changed": False, "diff": [], "response": [], "warnings": []}

    @abstractmethod
    def validate_input(self):
        if not self.config:
            self.msg = "config not available in playbook for validation"
            self.status = "failed"
            return self

    def get_diff_merged(self):
        # Implement logic to merge the resource configuration
        self.merged = True
        return self

    def get_diff_deleted(self):
        # Implement logic to delete the resource
        self.deleted = True
        return self

    def get_diff_replaced(self):
        # Implement logic to replace the resource
        self.replaced = True
        return self

    def get_diff_overridden(self):
        # Implement logic to overwrite the resource
        self.overridden = True
        return self

    def get_diff_gathered(self):
        # Implement logic to gather data about the resource
        self.gathered = True
        return self

    def get_diff_rendered(self):
        # Implement logic to render a configuration template
        self.rendered = True
        return self

    def get_diff_parsed(self):
        # Implement logic to parse a configuration file
        self.parsed = True
        return self

    def verify_diff_merged(self):
        # Implement logic to verify the merged resource configuration
        self.merged = True
        return self

    def verify_diff_deleted(self):
        # Implement logic to verify the deleted resource
        self.deleted = True
        return self

    def verify_diff_replaced(self):
        # Implement logic to verify the replaced resource
        self.replaced = True
        return self

    def verify_diff_overridden(self):
        # Implement logic to verify the overwritten resource
        self.overridden = True
        return self

    def verify_diff_gathered(self):
        # Implement logic to verify the gathered data about the resource
        self.gathered = True
        return self

    def verify_diff_rendered(self):
        # Implement logic to verify the rendered configuration template
        self.rendered = True
        return self

    def verify_diff_parsed(self):
        # Implement logic to verify the parsed configuration file
        self.parsed = True
        return self

    def setup_logger(self, logger_name):
        """Set up a logger with specified name and configuration based on dnac_log_level"""
        level_mapping = {
            'INFO': logging.INFO,
            'DEBUG': logging.DEBUG,
            'WARNING': logging.WARNING,
            'ERROR': logging.ERROR,
            'CRITICAL': logging.CRITICAL
        }
        level = level_mapping.get(self.dnac_log_level, logging.WARNING)

        logger = logging.getLogger(logger_name)
        # formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(module)s: %(funcName)s: %(lineno)d --- %(message)s', datefmt='%m-%d-%Y %H:%M:%S')
        formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s', datefmt='%m-%d-%Y %H:%M:%S')

        file_handler = logging.FileHandler(self.dnac_log_file_path, mode=self.dnac_log_mode)
        file_handler.setFormatter(formatter)

        logger.setLevel(level)
        logger.addHandler(file_handler)

    def validate_dnac_log_level(self):
        """Validates if the logging level is string and of expected value"""
        if self.dnac_log_level not in ('INFO', 'DEBUG', 'WARNING', 'ERROR', 'CRITICAL'):
            raise ValueError("Invalid log level: 'dnac_log_level:{0}'".format(self.dnac_log_level))

    def validate_dnac_log_file_path(self):
        """
        Validates the specified log file path, ensuring it is either absolute or relative,
        the directory exists, and has a .log extension.
        """
        # Convert the path to absolute if it's relative
        dnac_log_file_path = os.path.abspath(self.dnac_log_file_path)

        # Validate if the directory exists
        log_directory = os.path.dirname(dnac_log_file_path)
        if not os.path.exists(log_directory):
            raise FileNotFoundError("The directory for log file '{0}' does not exist.".format(dnac_log_file_path))

    def log(self, message, level="WARNING", frameIncrement=0):
        """Logs formatted messages with specified log level and incrementing the call stack frame
        Args:
            self (obj, required): An instance of the DnacBase Class.
            message (str, required): The log message to be recorded.
            level (str, optional): The log level, default is "info".
                                   The log level can be one of 'DEBUG', 'INFO', 'WARNING', 'ERROR', or 'CRITICAL'.
        """

        if self.dnac_log:
            # of.write("---- %s ---- %s@%s ---- %s \n" % (d, info.lineno, info.function, msg))
            # message = "Module: " + self.__class__.__name__ + ", " + message
            class_name = self.__class__.__name__
            callerframerecord = inspect.stack()[1 + frameIncrement]
            frame = callerframerecord[0]
            info = inspect.getframeinfo(frame)
            log_message = " %s: %s: %s: %s \n" % (class_name, info.function, info.lineno, message)
            log_method = getattr(self.logger, level.lower())
            log_method(log_message)

    def check_return_status(self):
        """API to check the return status value and exit/fail the module"""

        # self.log("status: {0}, msg:{1}".format(self.status, self.msg), frameIncrement=1)
        self.log("status: {0}, msg: {1}".format(self.status, self.msg), "DEBUG")
        if "failed" in self.status:
            self.module.fail_json(msg=self.msg, response=[])
        elif "exited" in self.status:
            self.module.exit_json(**self.result)
        elif "invalid" in self.status:
            self.module.fail_json(msg=self.msg, response=[])

    def is_valid_password(self, password):
        """
        Check if a password is valid.
        Args:
            self (object): An instance of a class that provides access to Cisco Catalyst Center.
            password (str): The password to be validated.
        Returns:
            bool: True if the password is valid, False otherwise.
        Description:
            The function checks the validity of a password based on the following criteria:
            - Minimum 8 characters.
            - At least one lowercase letter.
            - At least one uppercase letter.
            - At least one digit.
            - At least one special character
        """

        pattern = r"^(?=.*[a-z])(?=.*[A-Z])(?=.*\d)(?=.*[-=\\;,./~!@#$%^&*()_+{}[\]|:?]).{8,}$"

        return re.match(pattern, password) is not None

    def get_dnac_params(self, params):
        """Store the Cisco Catalyst Center parameters from the playbook"""

        dnac_params = {"dnac_host": params.get("dnac_host"),
                       "dnac_port": params.get("dnac_port"),
                       "dnac_username": params.get("dnac_username"),
                       "dnac_password": params.get("dnac_password"),
                       "dnac_verify": params.get("dnac_verify"),
                       "dnac_debug": params.get("dnac_debug"),
                       "dnac_log": params.get("dnac_log"),
                       "dnac_log_level": params.get("dnac_log_level"),
                       "dnac_log_file_path": params.get("dnac_log_file_path"),
                       "dnac_log_append": params.get("dnac_log_append")
                       }
        return dnac_params

    def get_task_details(self, task_id):
        """
        Get the details of a specific task in Cisco Catalyst Center.
        Args:
            self (object): An instance of a class that provides access to Cisco Catalyst Center.
            task_id (str): The unique identifier of the task for which you want to retrieve details.
        Returns:
            dict or None: A dictionary containing detailed information about the specified task,
            or None if the task with the given task_id is not found.
        Description:
            If the task with the specified task ID is not found in Cisco Catalyst Center, this function will return None.
        """

        result = None
        response = self.dnac._exec(
            family="task",
            function='get_task_by_id',
            params={"task_id": task_id}
        )

        self.log('Task Details: {0}'.format(str(response)), 'DEBUG')
        self.log("Retrieving task details by the API 'get_task_by_id' using task ID: {0}, Response: {1}".format(task_id, response), "DEBUG")

        if response and isinstance(response, dict):
            result = response.get('response')

        return result

    def check_task_response_status(self, response, validation_string, data=False):
        """
        Get the site id from the site name.

        Parameters:
            self - The current object details.
            response (dict) - API response.
            validation_string (string) - String used to match the progress status.

        Returns:
            self
        """

        if not response:
            self.msg = "response is empty"
            self.status = "exited"
            return self

        if not isinstance(response, dict):
            self.msg = "response is not a dictionary"
            self.status = "exited"
            return self

        response = response.get("response")
        if response.get("errorcode") is not None:
            self.msg = response.get("response").get("detail")
            self.status = "failed"
            return self

        task_id = response.get("taskId")
        while True:
            task_details = self.get_task_details(task_id)
            self.log('Getting task details from task ID {0}: {1}'.format(task_id, task_details), "DEBUG")

            if task_details.get("isError") is True:
                if task_details.get("failureReason"):
                    self.msg = str(task_details.get("failureReason"))
                else:
                    self.msg = str(task_details.get("progress"))
                self.status = "failed"
                break

            if validation_string in task_details.get("progress").lower():
                self.result['changed'] = True
                if data is True:
                    self.msg = task_details.get("data")
                self.status = "success"
                break

            self.log("progress set to {0} for taskid: {1}".format(task_details.get('progress'), task_id), "DEBUG")

        return self

    def reset_values(self):
        """Reset all neccessary attributes to default values"""

        self.have.clear()
        self.want.clear()

    def get_execution_details(self, execid):
        """
        Get the execution details of an API

        Parameters:
            execid (str) - Id for API execution

        Returns:
            response (dict) - Status for API execution
        """

        self.log("Execution Id: {0}".format(execid), "DEBUG")
        response = self.dnac._exec(
            family="task",
            function='get_business_api_execution_details',
            params={"execution_id": execid}
        )
        self.log("Response for the current execution: {0}".format(response))
        return response

    def check_execution_response_status(self, response):
        """
        Checks the reponse status provided by API in the Cisco Catalyst Center

        Parameters:
            response (dict) - API response

        Returns:
            self
        """

        if not response:
            self.msg = "response is empty"
            self.status = "failed"
            return self

        if not isinstance(response, dict):
            self.msg = "response is not a dictionary"
            self.status = "failed"
            return self

        executionid = response.get("executionId")
        while True:
            execution_details = self.get_execution_details(executionid)
            if execution_details.get("status") == "SUCCESS":
                self.result['changed'] = True
                self.msg = "Successfully executed"
                self.status = "success"
                break

            if execution_details.get("bapiError"):
                self.msg = execution_details.get("bapiError")
                self.status = "failed"
                break

        return self

    def check_string_dictionary(self, task_details_data):
        """
        Check whether the input is string dictionary or string.

        Parameters:
            task_details_data (string) - Input either string dictionary or string.

        Returns:
            value (dict) - If the input is string dictionary, else returns None.
        """

        try:
            value = json.loads(task_details_data)
            if isinstance(value, dict):
                return value
        except json.JSONDecodeError:
            pass
        return None

    def camel_to_snake_case(self, config):
        """
        Convert camel case keys to snake case keys in the config.

        Parameters:
            config (list) - Playbook details provided by the user.

        Returns:
            new_config (list) - Updated config after eliminating the camel cases.
        """

        if isinstance(config, dict):
            new_config = {}
            for key, value in config.items():
                new_key = re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', key).lower()
                if new_key != key:
                    self.log("{0} will be deprecated soon. Please use {1}.".format(key, new_key), "DEBUG")
                new_value = self.camel_to_snake_case(value)
                new_config[new_key] = new_value
        elif isinstance(config, list):
            return [self.camel_to_snake_case(item) for item in config]
        else:
            return config
        return new_config

    def update_site_type_key(self, config):
        """
        Replace 'site_type' key with 'type' in the config.

        Parameters:
            config (list or dict) - Configuration details.

        Returns:
            updated_config (list or dict) - Updated config after replacing the keys.
        """

        if isinstance(config, dict):
            new_config = {}
            for key, value in config.items():
                if key == "site_type":
                    new_key = "type"
                else:
                    new_key = re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', key).lower()
                new_value = self.update_site_type_key(value)
                new_config[new_key] = new_value
        elif isinstance(config, list):
            return [self.update_site_type_key(item) for item in config]
        else:
            return config

        return new_config


def is_list_complex(x):
    return isinstance(x[0], dict) or isinstance(x[0], list)


def has_diff_elem(ls1, ls2):
    return any((elem not in ls1 for elem in ls2))


def compare_list(list1, list2):
    len_list1 = len(list1)
    len_list2 = len(list2)
    if len_list1 != len_list2:
        return False

    if len_list1 == 0:
        return True

    attempt_std_cmp = list1 == list2
    if attempt_std_cmp:
        return True

    if not is_list_complex(list1) and not is_list_complex(list2):
        return set(list1) == set(list2)

    # Compare normally if it exceeds expected size * 2 (len_list1==len_list2)
    MAX_SIZE_CMP = 100
    # Fail fast if elem not in list, thanks to any and generators
    if len_list1 > MAX_SIZE_CMP:
        return attempt_std_cmp
    else:
        # not changes 'has diff elem' to list1 != list2 ':lists are not equal'
        return not (has_diff_elem(list1, list2)) or not (has_diff_elem(list2, list1))


def fn_comp_key(k, dict1, dict2):
    return dnac_compare_equality(dict1.get(k), dict2.get(k))


def dnac_compare_equality(current_value, requested_value):
    # print("dnac_compare_equality", current_value, requested_value)
    if requested_value is None:
        return True
    if current_value is None:
        return True
    if isinstance(current_value, dict) and isinstance(requested_value, dict):
        all_dict_params = list(current_value.keys()) + list(requested_value.keys())
        return not any((not fn_comp_key(param, current_value, requested_value) for param in all_dict_params))
    elif isinstance(current_value, list) and isinstance(requested_value, list):
        return compare_list(current_value, requested_value)
    else:
        return current_value == requested_value


def simple_cmp(obj1, obj2):
    return obj1 == obj2


def get_dict_result(result, key, value, cmp_fn=simple_cmp):
    if isinstance(result, list):
        if len(result) == 1:
            if isinstance(result[0], dict):
                result = result[0]
                if result.get(key) is not None and result.get(key) != value:
                    result = None
            else:
                result = None
        else:
            for item in result:
                if isinstance(item, dict) and (item.get(key) is None or item.get(key) == value):
                    result = item
                    return result
            result = None
    elif not isinstance(result, dict):
        result = None
    elif result.get(key) is not None and result.get(key) != value:
        result = None
    return result


def dnac_argument_spec():
    argument_spec = dict(
        dnac_host=dict(type="str", required=True),
        dnac_port=dict(type="int", required=False, default=443),
        dnac_username=dict(type="str", default="admin", aliases=["user"]),
        dnac_password=dict(type="str", no_log=True),
        dnac_verify=dict(type="bool", default=True),
        dnac_version=dict(type="str", default="2.2.3.3"),
        dnac_debug=dict(type="bool", default=False),
        validate_response_schema=dict(type="bool", default=True),
    )
    return argument_spec


def validate_str(item, param_spec, param_name, invalid_params):
    """
    This function checks that the input `item` is a valid string and confirms to
    the constraints specified in `param_spec`. If the string is not valid or does
    not meet the constraints, an error message is added to `invalid_params`.

    Args:
        item (str): The input string to be validated.
        param_spec (dict): The parameter's specification, including validation constraints.
        param_name (str): The name of the parameter being validated.
        invalid_params (list): A list to collect validation error messages.

    Returns:
        str: The validated and possibly normalized string.

    Example `param_spec`:
        {
            "type": "str",
            "length_max": 255  # Optional: maximum allowed length
        }
    """

    item = validation.check_type_str(item)
    if param_spec.get("length_max"):
        if 1 <= len(item) <= param_spec.get("length_max"):
            return item
        else:
            invalid_params.append(
                "{0}:{1} : The string exceeds the allowed "
                "range of max {2} char".format(param_name, item, param_spec.get("length_max"))
            )
    return item


def validate_int(item, param_spec, param_name, invalid_params):
    """
    This function checks that the input `item` is a valid integer and conforms to
    the constraints specified in `param_spec`. If the integer is not valid or does
    not meet the constraints, an error message is added to `invalid_params`.

    Args:
        item (int): The input integer to be validated.
        param_spec (dict): The parameter's specification, including validation constraints.
        param_name (str): The name of the parameter being validated.
        invalid_params (list): A list to collect validation error messages.

    Returns:
        int: The validated integer.

    Example `param_spec`:
        {
            "type": "int",
            "range_min": 1,     # Optional: minimum allowed value
            "range_max": 100    # Optional: maximum allowed value
        }
    """

    item = validation.check_type_int(item)
    min_value = 1
    if param_spec.get("range_min") is not None:
        min_value = param_spec.get("range_min")
    if param_spec.get("range_max"):
        if min_value <= item <= param_spec.get("range_max"):
            return item
        else:
            invalid_params.append(
                "{0}:{1} : The item exceeds the allowed "
                "range of max {2}".format(param_name, item, param_spec.get("range_max"))
            )
    return item


def validate_bool(item, param_spec, param_name, invalid_params):
    """
    This function checks that the input `item` is a valid boolean value. If it does
    not represent a valid boolean value, an error message is added to `invalid_params`.

    Args:
        item (bool): The input boolean value to be validated.
        param_spec (dict): The parameter's specification, including validation constraints.
        param_name (str): The name of the parameter being validated.
        invalid_params (list): A list to collect validation error messages.

    Returns:
        bool: The validated boolean value.
    """

    return validation.check_type_bool(item)


def validate_list(item, param_spec, param_name, invalid_params):
    """
    This function checks if the input `item` is a valid list based on the specified `param_spec`.
    It also verifies that the elements of the list match the expected data type specified in the
    `param_spec`. If any validation errors occur, they are appended to the `invalid_params` list.

    Args:
        item (list): The input list to be validated.
        param_spec (dict): The parameter's specification, including validation constraints.
        param_name (str): The name of the parameter being validated.
        invalid_params (list): A list to collect validation error messages.

    Returns:
        list: The validated list, potentially normalized based on the specification.
    """

    try:
        if param_spec.get("type") == type(item).__name__:
            keys_list = []
            for dict_key in param_spec:
                keys_list.append(dict_key)
            if len(keys_list) == 1:
                return validation.check_type_list(item)

            temp_dict = {keys_list[1]: param_spec[keys_list[1]]}
            try:
                if param_spec['elements']:
                    get_spec_type = param_spec['type']
                    get_spec_element = param_spec['elements']
                    if type(item).__name__ == get_spec_type:
                        for element in item:
                            if type(element).__name__ != get_spec_element:
                                invalid_params.append(
                                    "{0} is not of the same datatype as expected which is {1}".format(element, get_spec_element)
                                )
                    else:
                        invalid_params.append(
                            "{0} is not of the same datatype as expected which is {1}".format(item, get_spec_type)
                        )
            except Exception as e:
                item, list_invalid_params = validate_list_of_dicts(item, temp_dict)
                invalid_params.extend(list_invalid_params)
        else:
            invalid_params.append("{0} : is not a valid list".format(item))
    except Exception as e:
        invalid_params.append("{0} : comes into the exception".format(e))

    return item


def validate_dict(item, param_spec, param_name, invalid_params):
    """
    This function checks if the input `item` is a valid dictionary based on the specified `param_spec`.
    If the dictionary does not match the expected data type specified in the `param_spec`,
    a validation error is appended to the `invalid_params` list.

    Args:
        item (dict): The input dictionary to be validated.
        param_spec (dict): The parameter's specification, including validation constraints.
        param_name (str): The name of the parameter being validated.
        invalid_params (list): A list to collect validation error messages.

    Returns:
        dict: The validated dictionary.
    """

    if param_spec.get("type") != type(item).__name__:
        invalid_params.append("{0} : is not a valid dictionary".format(item))
    return validation.check_type_dict(item)


def validate_list_of_dicts(param_list, spec, module=None):
    """Validate/Normalize playbook params. Will raise when invalid parameters found.
    param_list: a playbook parameter list of dicts
    spec: an argument spec dict
          e.g. spec = dict(ip=dict(required=True, type='bool'),
                           foo=dict(type='str', default='bar'))
    return: list of normalized input data
    """

    v = validation
    normalized = []
    invalid_params = []

    for list_entry in param_list:
        valid_params_dict = {}
        if not spec:
            # Handle the case when spec becomes empty but param list is still there
            invalid_params.append("No more spec to validate, but parameters remain")
            break
        for param in spec:
            item = list_entry.get(param)
            if item is None:
                if spec[param].get("required"):
                    invalid_params.append(
                        "{0} : Required parameter not found".format(param)
                    )
                else:
                    item = spec[param].get("default")
                    valid_params_dict[param] = item
                    continue
            data_type = spec[param].get("type")
            switch = {
                "str": validate_str,
                "int": validate_int,
                "bool": validate_bool,
                "list": validate_list,
                "dict": validate_dict,
            }

            validator = switch.get(data_type)
            if validator:
                item = validator(item, spec[param], param, invalid_params)
            else:
                invalid_params.append(
                    "{0}:{1} : Unsupported data type {2}.".format(param, item, data_type)
                )

            choice = spec[param].get("choices")
            if choice:
                if item not in choice:
                    invalid_params.append(
                        "{0} : Invalid choice provided".format(item)
                    )

            no_log = spec[param].get("no_log")
            if no_log:
                if module is not None:
                    module.no_log_values.add(item)
                else:
                    msg = "\n\n'{0}' is a no_log parameter".format(param)
                    msg += "\nAnsible module object must be passed to this "
                    msg += "\nfunction to ensure it is not logged\n\n"
                    raise Exception(msg)

            valid_params_dict[param] = item
        normalized.append(valid_params_dict)

    return normalized, invalid_params


class DNACSDK(object):
    def __init__(self, params):
        self.result = dict(changed=False, result="")
        self.validate_response_schema = params.get("validate_response_schema")
        if DNAC_SDK_IS_INSTALLED:
            self.api = api.DNACenterAPI(
                username=params.get("dnac_username"),
                password=params.get("dnac_password"),
                base_url="https://{dnac_host}:{dnac_port}".format(
                    dnac_host=params.get("dnac_host"), dnac_port=params.get("dnac_port")
                ),
                version=params.get("dnac_version"),
                verify=params.get("dnac_verify"),
                debug=params.get("dnac_debug"),
            )
            if params.get("dnac_debug") and LOGGING_IN_STANDARD:
                logging.getLogger('dnacentersdk').addHandler(logging.StreamHandler())
        else:
            self.fail_json(msg="DNA Center Python SDK is not installed. Execute 'pip install dnacentersdk'")

    def changed(self):
        self.result["changed"] = True

    def object_created(self):
        self.changed()
        self.result["result"] = "Object created"

    def object_updated(self):
        self.changed()
        self.result["result"] = "Object updated"

    def object_deleted(self):
        self.changed()
        self.result["result"] = "Object deleted"

    def object_already_absent(self):
        self.result["result"] = "Object already absent"

    def object_already_present(self):
        self.result["result"] = "Object already present"

    def object_present_and_different(self):
        self.result["result"] = "Object already present, but it has different values to the requested"

    def object_modify_result(self, changed=None, result=None):
        if result is not None:
            self.result["result"] = result
        if changed:
            self.changed()

    def is_file(self, file_path):
        return os.path.isfile(file_path)

    def extract_file_name(self, file_path):
        return os.path.basename(file_path)

    def _exec(self, family, function, params=None, op_modifies=False, **kwargs):
        try:
            family = getattr(self.api, family)
            func = getattr(family, function)
        except Exception as e:
            self.fail_json(msg=e)

        try:
            if params:
                file_paths_params = kwargs.get('file_paths', [])
                # This substitution is for the import file operation
                if file_paths_params and isinstance(file_paths_params, list):
                    multipart_fields = {}
                    for (key, value) in file_paths_params:
                        if isinstance(params.get(key), str) and self.is_file(params[key]):
                            file_name = self.extract_file_name(params[key])
                            file_path = params[key]
                            multipart_fields[value] = (file_name, open(file_path, 'rb'))

                    params.setdefault("multipart_fields", multipart_fields)
                    params.setdefault("multipart_monitor_callback", None)

                if not self.validate_response_schema and op_modifies:
                    params["active_validation"] = False

                response = func(**params)
            else:
                response = func()
        except exceptions.dnacentersdkException as e:
            self.fail_json(
                msg=(
                    "An error occured when executing operation."
                    " The error was: {error}"
                ).format(error=to_native(e))
            )
        return response

    def fail_json(self, msg, **kwargs):
        self.result.update(**kwargs)
        raise Exception(msg)

    def exit_json(self):
        return self.result


def main():
    pass


if __name__ == "__main__":
    main()
